type numbering = array [vertex] of vertex;

procedure StrongComponents(var (* for speed *) G: graph;
                           var components: numbering);
  (* The procedure finds all strongly connected components in a directed graph;
     it labels all vertices in a given strongly connected component
     (in components[ ]) by the name of the component's root, which is an
     internally chosen representative for the strongly connected component. *)

  var dfsnumber, highwater: array [vertex] of 0 .. MaxVertex;
      onstack: array [vertex] of boolean;
      lastdfsnumber: 0 .. MaxVertex;
      i: vertex;

  (* Program assumes a stack package with operations
       procedure CreateStack, procedure Push(v: vertex),
       and procedure Pop(var v: vertex). *)

  procedure Strong(v: vertex);
    var p: PtrToNode;
        w: vertex;
    begin
      (* Set up bookkeeping for v. *)
      lastdfsnumber := lastdfsnumber + 1;
      dfsnumber[v] := lastdfsnumber;
      highwater[v] := lastdfsnumber;
      Push(v); onstack[v] := true;

      (* Traverse the adjacency list of v carrying on the depth-first search. *)
      p := G.AdjLists[v];
      while p <> nil do
        begin
          w := p^.id;
          if dfsnumber[w] = 0
            then begin
                   (* (v,w) is a tree arc, continue the depth-first search. *)
                   Strong(w);
                   (* On return, update highwater of parent as needed. *)
                   if highwater[w] < highwater[v]
                     then highwater[v] := highwater[w]
                 end
          else if (* if (v,w) is a back arc or a relevant cross arc and the
                     end vertex improves the highwater mark of the parent... *)
                  (dfsnumber[w] < dfsnumber[v]) and onstack[w] and
                  (dfsnumber[w] < highwater[v])
            then highwater[v] := dfsnumber[w];
          p := p^.next
        end;
      (* We are ready to leave vertex v, but must first check whether
         or not v is the root of a strongly connected component. *)
      if highwater[v] = dfsnumber[v]
        then repeat (* Pop stack until the root of the component is popped. *)
               Pop(w);
               components[w] := v;
               onstack[w] := false
             until w = v
    end; (* Strong *)

  begin
    (* Find all the strongly connected components. *)
    CreateStack;
    lastdfsnumber := 0;
    for i := 1 to G.size do dfsnumber[i] := 0;
    for i := 1 to G.size do
      if dfsnumber[i] = 0 then Strong(i)
  end; (* StrongComponents *)
